import wx
import wx.richtext
import threading
import time
import os

#TODO: Make this use fewer resources
#TODO: Reload files that are modified on disk

class HighlighterPanel(wx.Panel):
    """An extension of wxPanel for highlighting Python code. HighlighterPanels are intended to be pages in wxNotebook objects."""
    PANEL_COUNT = 0
    
    highlight_attrs = []
    for i in range(256):
        t = wx.richtext.TextAttrEx()
        t.SetBackgroundColour(wx.Colour(255,255,i))
        highlight_attrs.append(t)
    special_attr = wx.richtext.TextAttrEx()
    special_attr.SetBackgroundColour(wx.Colour(255,255,0))
    
    def __init__(self, parent,text_file_path,notebook,img_count):
        """text_file_path should be the path to the .py file that is being highlighted.
notebook should be the wxNotebook in which this panel is a page.
img_count should be the length of the ImageList containing tab images in the notebook."""

        #Call the parent contstructor
        wx.Panel.__init__(self,parent)

        #Open and read the text file
        text_file = open(text_file_path, "r")
        self.last_modified = os.stat(text_file_path).st_mtime
        self.text = text_file.read()
        #Kill non-\n newlines
        self.text = self.text.replace("\r\n","\n")
        self.text = self.text.replace("\r","\n")
        #Split the text into lines
        self.text_lines = self.text.splitlines()

        #Create the list of highlight colours for each row, defaulting to white
        self.line_colours = [255 for i in range(len(self.text_lines))]

        #Create and populate the list of line offsets, since to highlight
        #with wxTextCtrl we need to know the start and end offsets in chars
        self.line_offsets = []
        count = 0
        for line in self.text_lines:
            self.line_offsets.append(count)
            count += len(line)+1

        #Create the wxTextCtrl that holds all the text, using the multiline, rich
        #text, vertical scroll, and read-only flags
        self.text_box = wx.richtext.RichTextCtrl(self, -1, self.text, style=wx.TE_MULTILINE|wx.TE_RICH|wx.VSCROLL|wx.TE_READONLY)

        #Create a wxBoxSizer, add the text box to it in expand mode, and
        #attach the sizer to this panel
        sizer = wx.BoxSizer()
        sizer.Add(self.text_box,1,wx.EXPAND)
        self.SetSizer(sizer)

        #Create a timer to fire off every 1/10th of a second that will
        #cause the highlighting to be refreshed
        self.timer = wx.Timer(self)
        self.timer.Start(1000/10)
        self.Bind(wx.EVT_TIMER, self.OnTimer)

        #Store or initialise other variables
        self.cur_img_index = 0
        self.last_img_index = 0
        self.img_count = img_count
        self.notebook = notebook
        self.touched = {}
        self.is_paused = True
        self.text_file_path = text_file_path
        self.do_step = False
        self.pipe = None
        self.is_alive = True
        self.was_touched = False
        
        #Increment the static count
        HighlighterPanel.PANEL_COUNT += 1
        
        self.most_recent = -1
        self.most_recent_old = -1
        
    def set_pipe(self,pipe):
        self.die()
        
        self.is_alive = True
        self.pipe = pipe
        
        #Set up the thread to watch the pipe
        self.pipe_thread = threading.Thread(target=self.do_thread_stuff)
        self.pipe_thread.start()

    def pause(self):
        """Stops this panel from updating its contents."""
        self.is_paused = True

    def die(self):
        self.is_alive = False
        try:
            self.pipe_thread.join(10)
        except:
            pass
        if self.pipe != None and not self.pipe.closed:
            self.pipe.send("rm")

    def resume(self):
        """Resumes the process of updating this panel's contents."""
        self.is_paused = False

    def OnLineTouched(self, line_num):
        """Registers that a line was touched for handling at the next refresh."""
        self.touched[line_num] = True
        self.was_touched = True
        self.resume()

    def do_thread_stuff(self):
        while self.is_alive:
            if self.pipe != None:
                try:
                    if self.pipe.poll(0.1):
                        msg = self.pipe.recv()
                        if type(msg) == int:
                            self.OnLineTouched(msg)
                except (IOError, EOFError): #IOError will occur if the pipe has been closed on the other side
                    self.pipe = None
            else:
                time.sleep(0.1)
        
    def OnTimer(self, event):
        """This function is called by the timer set in the constructor.
It processes the list of touched lines and updates the display accordingly if this panel is not paused."""
        
        if (not self.is_paused and self.is_alive) or self.do_step:
            self.do_step = False
            #if self.IsShown():
            for line_num in range(len(self.text_lines)):
                old = self.line_colours[line_num]
                if line_num in self.touched.keys():
                    #If the line was touched, move toward yellow quickly
                    self.line_colours[line_num] = self.line_colours[line_num]-100
                    if self.line_colours[line_num] < 0:
                        self.line_colours[line_num] = 0
                    self.most_recent_old = self.most_recent
                    self.most_recent = line_num
                elif self.line_colours[line_num] != 255:
                    #Otherwise, move back toward white
                    self.line_colours[line_num] = self.line_colours[line_num]+25
                    if self.line_colours[line_num] > 255:
                        self.line_colours[line_num] = 255
                if self.line_colours[line_num] != old:
                    #Only modify the style if the colour actually changed
                    if self.most_recent == line_num:
                        ta = HighlighterPanel.special_attr
                        self.text_box.SetStyle([self.line_offsets[self.most_recent_old],self.line_offsets[self.most_recent_old]+len(self.text_lines[self.most_recent_old])],HighlighterPanel.highlight_attrs[self.line_colours[self.most_recent_old]])
                    else:
                        ta = HighlighterPanel.highlight_attrs[self.line_colours[line_num]]
                    self.text_box.SetStyle([self.line_offsets[line_num],self.line_offsets[line_num]+len(self.text_lines[line_num])],ta)
            
            if self.touched:
                #If any lines were touched, intensify the status light in this panel's tab in the notebook
                self.cur_img_index += 8
                if self.cur_img_index > self.img_count-1:
                    self.cur_img_index = self.img_count-1
                self.was_touched = False
            else:
                #Otherwise, turn down the status light's intensity
                self.cur_img_index = self.cur_img_index-4
                if self.cur_img_index < 0:
                    self.cur_img_index = 0
            if self.last_img_index != self.cur_img_index:
                self.last_img_index = self.cur_img_index
                self.notebook.SetPageImageByReference(self,self.cur_img_index)
            
            #Clear out the list of touched lines
            self.touched.clear()

    def step(self):
        self.do_step = True
